#
#
#

""" Make unfolded starting structures """
__author__ = "Magdalena A. Jonikas"
__version__ = "1.0"
__date__ = " 7/15/09 "

import sys
import simtk.nast.molecule as molecule
import simtk.nast.simplePdb as pdb
import simtk.nast.defaultNastParameters as params
import random
import math
import time

numSteps        = 50000
stepsPerReport      = 1000
defaultTemperature  = 300.0
randomSeed      = 1
tryUsingGpu          = False
verbose         = True

def calcDist(pos1,pos2):
    (x1,y1,z1)=pos1
    (x2,y2,z2)=pos2
    distance = ((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)**0.5
    return distance 

def arange(start, stop=None, step=None):
    list = []
    if stop is None:
        stop = float(start)
        start = 0.0
    if step is None:
        step = 1.0
    cur = float(start)
    while cur< stop:
        list.append(cur)
        cur+=step
    return list

class BuildMol(object):
    def __init__(self, xtalIn, ssIn):
        # Make a molecule object from input crystal structure and secondary structure
        self.m0 = molecule.Molecule(pdbFilename=xtalIn, helixFilename=ssIn)

    def buildNewPdbString(self, currentString):
        self.addLeft = True
        self.addRight = True
        newString = ''
        if self.step<self.leftSide:
            newString+=self.m0.pdb.originalPdbLines[self.startRes-(self.step+1)]  + '\n'
        else: 
            self.addLeft = False
            #print "done with left side"
        newString+=currentString+'\n'
        if self.step<self.rightSide:
            newString+=self.m0.pdb.originalPdbLines[self.startRes+self.step+1] + '\n'
        else: 
            self.addRight = False
            #print "done with right side"
        if not self.addLeft and not self.addRight:
            #print "nothing to add on either end, need to stop this somehow, returning input string"
            return currentString
        return newString

    def setupBuild(self):
        # Setup the building process by choosing a starting point
        self.startRes = random.randrange(self.m0.numParticles)
        self.trace=[]
        # Figure out how many residues need to be added on each side
        self.leftSide = self.startRes
        self.rightSide = self.m0.numParticles-1-self.startRes
        self.step = 0

        startPdbString = self.buildNewPdbString(self.m0.pdb.originalPdbLines[self.startRes])
        startPdb = pdb.SimplePdb(pdbString=startPdbString)
        
        # Make starting positions for starting residues
        startSequence=self.m0.pdb.data[self.startRes-1]['resName']+'\n'\
                    + self.m0.pdb.data[self.startRes]['resName']+'\n'\
                    + self.m0.pdb.data[self.startRes+1]['resName']+'\n'
        newPosPdb = pdb.SimplePdb(sequenceString=startSequence)
        newPos = newPosPdb.getPositions()
        startPdb.setPositions(newPos)
        startHelixData = self.keepReleventHelices(startPdb)
        self.nextPdbString = self.runMD(startPdb, startHelixData)
        #print "setup result:"
        #print self.nextPdbString
        self.trace.append(self.nextPdbString)
        #print "finished setup"

    def runAddStep(self, startString):
        currentPdb = self.addNextRes(startString)
        if currentPdb==None:
            return None
        if currentPdb.getPdb() == startString:
            return startString
        currentString = currentPdb.getPdb()
        currentPdb = self.addNextRes(currentString)
        if currentPdb==None:
            return None
        currentHelixData = self.keepReleventHelices(currentPdb)
        nextPdbString = self.runMD(currentPdb, currentHelixData)
        return nextPdbString

    def keepReleventHelices(self, newPdb):
        offset = newPdb.data[0]['resNum']
        resNumList = []
        newLength = len(newPdb.data)
        for resId in range(newLength):
            resNumList.append(newPdb.data[resId]['resNum'])
        #print resNumList
        helixLists = []
        helixBasePairs = {}
        helixNum = {}
        helixID=0
        for helix in self.m0.helixLists:
            newHelix = ([],[])
            for i in range(len(helix[0])):
                resa = helix[0][i]
                resb = helix[1][i]
                if resa in resNumList and resb in resNumList:
                    resi = resa-offset
                    resj = resb-offset
                    newHelix[0].append(resi)
                    newHelix[1].append(resj)
                    helixBasePairs[resi]=resj
                    helixBasePairs[resj]=resi
                    helixNum[resi]=helixID
                    helixNum[resj]=helixID
            if len(newHelix[0])>0:
                helixLists.append(newHelix)
                helixID+=1
        #print helixLists
        return (helixLists, helixBasePairs, helixNum)

    def runMD(self, newPdb, newHelixData):
        m=molecule.Molecule( pdbObject = newPdb,
                       helixData = newHelixData,
                       defaultTemperature=defaultTemperature,
                       randomSeed=randomSeed,
                       tryUsingGpu=tryUsingGpu)
        #fEnergyOut=sys.stdout
        fEnergyOut=None
        m.doSimulation(numSteps=numSteps,
                    stepsPerReport=stepsPerReport,
                    fEnergyOut=fEnergyOut)

        (t, eK, eP, p, v, f) =  m.context.getState(getPositions=True,
                                                  getVelocities=False,
                                                  getForces=False)
        #remarkString='REMARK  STEP %d\n' % step
        #remarkString+='REMARK  ENERGIES: EK = %0.1f EP = %0.1f' % (eK, eP)
        #remarkString='REMARK  ENERGIES: EK = %0.1f EP = %0.1f' % (eK, eP)
        s=m.pdb.getPdb(positions=p, useOriginalPdb=False)
        #fPdbOut = sys.stdout
        #fPdbOut.write('%s\n%sEND\n' % (remarkString,s))
        return s

    def addNextRes(self, currentString):
        self.step+=1
        currentN = len(currentString)
        newPdbString = self.buildNewPdbString(currentString)
        newPdb = pdb.SimplePdb(pdbString=newPdbString)
        oldPos = newPdb.getPositions()
        newPos = oldPos
        if self.addLeft:
            leftPosAnchor = oldPos[1]
            leftPos = self.chooseNewPos(leftPosAnchor, oldPos[1:])
            if leftPos==None:
                return None
            newPos[0] = leftPos
        if self.addRight:
            rightPosAnchor = oldPos[-2]
            rightPos = self.chooseNewPos(rightPosAnchor, oldPos[:-1])
            if rightPos==None:
                return None
            newPos[-1] = rightPos
        newPdb.setPositions(newPos)
        return newPdb
        #print newPdb.getPdb()
        
    def chooseNewPos(self, lastPos, oldPos):
        (lastx,lasty,lastz) = lastPos
        R = params.bondR_bb
        zList = arange(-1*R, R, R/20.0)
        phiList = arange(0,2*math.pi, 0.3)
        possibleNewPositions = []
        for z in zList:
            theta = math.asin(z/R)
            for phi in phiList:
                x = R*math.cos(theta)*math.cos(phi)
                y = R*math.cos(theta)*math.sin(phi)
                newx = lastx+x
                newy = lasty+y
                newz = lastz+z
                possibleNewPositions.append( (newx,newy,newz) )
        nearbyPositions = []
        refPos=oldPos[1]
        for i in range(len(oldPos)):
            checkPos = oldPos[i]
            dist = calcDist(refPos, checkPos)
            if dist<(2*R):
                nearbyPositions.append(checkPos)
        goodPositions=[]
        for position in possibleNewPositions:
            for nearbyPos in nearbyPositions:
                dist = calcDist(position, nearbyPos)
                if dist>R:
                    goodPositions.append(position)
        if len(goodPositions)<1:
            return None
        #print len(possibleNewPositions)
        #print len(goodPositions)
        chosenPosition = random.choice(goodPositions)
        return chosenPosition
    

if __name__=='__main__':
    t0=time.time()
    xtalIn = sys.argv[1]
    ssIn = sys.argv[2]
    id = int(sys.argv[3])
    rootname=xtalIn[0:4]

    b = BuildMol(xtalIn, ssIn)
    b.setupBuild()
    startString = b.nextPdbString
    step=0
    b.errorCount=0
    while startString:
        nextString = b.runAddStep(startString)
        if nextString==None:
            b.errorCount+=1
            if b.errorCount>5: break
            if len(b.trace)>(b.errorCount*2):
                newTrace = b.trace[0:(-2*b.errorCount)]
            else: newTrace = []
            nextString=newTrace[-1]
            b.trace = newTrace
        if nextString == startString:
            break
        b.trace.append(nextString)
        startString = nextString
        pdbOut = open('frame-%i.pdb' % step, 'w')
        pdbOut.writelines(nextString+'END\n')
        pdbOut.close()
        step+=1
        #if b.step>5: break

    lastPdb=pdb.SimplePdb(pdbString=nextString)
    lastHelixData = b.keepReleventHelices(lastPdb)
    lastPdbString = lastPdb.getPdb()
    #lastPdbString = b.runMD(lastPdb, lastHelixData)
    #lastPdb=pdb.SimplePdb(pdbString=lastPdbString)
    #lastPdbString = b.runMD(lastPdb, lastHelixData)
    #lastPdb=pdb.SimplePdb(pdbString=lastPdbString)

    tEnd = time.time()
    rLine = 'REMARK TIME %.2f\n' % (tEnd-t0)
    pdbOut = open('%s-%i.pdb' % (rootname, id), 'w')
    pdbOut.writelines(rLine+lastPdbString+'END\n')
    pdbOut.close()
    #psfOut = open('last-frame.psf', 'w')
    #psfOut.writelines(lastPdb.getPsf())
    #psfOut.close()
